/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Michael Kristensen, mik@caffrey.dk]          *
***********************************************************
*  -- History --                                          *
*    release 0.1 - 06-03-2003                             *
*            http://caffrey.dk/megasquirt                 *
\*********************************************************/

   /*!
   * \addtogroup g_idle_air Idle Air Control Routines
   */
/*@{*/

#include "global.h"
#include "iac.h"
#include "helpers.h"
#include "adc.h"
#include "ve.h"
#include "fuelcalc.h"
#include "benchmark.h"
#include "mainloop.h"

#include <inttypes.h>
#include <avr/io.h>
#include <stdlib.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "arithm.c"
#include "lbus.c"
#include "softpwm.c"
#include "injector.c"
#if A340E
#include "a340e.h"
#endif

struct iac_t iac;

/** originally iac.shared.resync = 10 * config.iac_overclose_interval
 */
/*void iac_resync_reset( void )
{
	iac.shared.stepperonly.resync = 255;
}
*/
/* Resync the stepper iac to correct any drift.
 * Questionable if necessary at all...
 * It likely never worked as intended, as it was called from branch
 * tps < iac_tps_threshold && error outside deadband
 * where iac_position is unlikely to be 0
 * caller must make sure that stepper iac is configured
 */
/*static inline uint8_t stepper_time_to_overclose( void )
{
	if (iac.shared.stepperonly.resync != 0) {
		iac.shared.stepperonly.resync--;
		return 0; // not yet !
	}

	iac_resync_reset();
	// time for overclose:
	return 1; 
}
*/
/**
 * Sets iac solenoid duty without VBATT scaling
 * duty will be available in idle_pwm_soft.designated_ratio (but who cares?)
 */
static inline void set_iac_solenoid_duty( uint8_t duty )
{
#ifndef TRUE_PWM_IAC // new experimental
	softpwm_set(&iac.shared.idle_pwm_soft, duty);
#endif
	// iac.position = duty; INTENTIONALLY NOT !
}

/**
 * Only call for solenoid iac (not for stepper)
 * From a position of 90 (/256), it will take 45 * 100msec (=4.5 seconds)
 * to unpower iac. Note that the Audi Bosch iac actually opens
 * when going from 25% to 0% PWM duty. No big deal.
 * calc_pwm_dc() VBATT scaleup omitted.
 *
 * Sami proposed a bit higher threshold than
 * engine.tps > actual_tps_thres (to avoid stall)
 * let it be 4% TPS (well enough air):
 */
#define NORAMPDOWN_TPSWINDOW 10
// another TPS threshold (for tpsfollow and PID weight):
#define NOEXTRAAIR_TPSWINDOW 16
#define SLOWLY_CLOSE_STEPS 2
void slowly_close_solenoid( void )
{
	uint8_t tps_thresh = actual_tps_thres;
	if( tps_thresh < 255-NORAMPDOWN_TPSWINDOW ){
		tps_thresh += NORAMPDOWN_TPSWINDOW;
		if( engine.tps <= tps_thresh ){ // don't ramp down duty (any further)
			return;
		}
	}
	uint8_t pos = iac.position;
	if( pos > SLOWLY_CLOSE_STEPS ){
		pos -= SLOWLY_CLOSE_STEPS;
	} else {
		pos = 0;
	}
	iac.position = pos;
	set_iac_solenoid_duty( pos );
	// set_iac_pos( pos ); would apply VBATT, but really superfluous for this
};

/** ================= set idle bit in engine.status
 */
void in_idle( void )
{
	if (engine.status & _BV(idle)){	// was already in idle
		return;
	}
	// just entered idle mode:

	iac.closing_delay = 10 * config.iac_close_delay;	//#1.1 reset counter..

	cli();
	engine.status |= _BV(idle);
	sei();
}

/** ================= reset idle bit in engine.status
 * Also resets the PID integral
 */
void not_in_idle( void )
{

// #1.1 zero integral, and start close solenoid slowly(if not tps_follow_configured) after a while !


	if (iac.closing_delay) {
		iac.closing_delay--;
	} else {
		iac.integral = 0; // this one IS necessary!
		if(iac_solenoid_configured() && !(iac_tps_follow_configured()))
			slowly_close_solenoid();
	}

//---#

	if((engine.status & _BV(idle)) == 0) {	// wasn't in idle anyway
		return;
	}

	// just left idle mode:

	if(iac_stepper_configured()&&(config_ext.iac_conf2&_BV(float_stepper_pos))) {
		iac.position = lookup_iac_clt_ref_pos();// cheat for stepper sync
	}

	cli();
	engine.status &= ~_BV(idle);
	sei();

	iac.rpmdot = 0;
	iac.pid = 0;
	iac.pid_p = 0;
	iac.pid_i = 0;
	iac.pid_d = 0;
}
/** ===================
* Final RPM target for actual engine.coolant
*/
uint8_t lookup_clt_rpmtarget( void )
{
	uint8_t	final_rpm = linear_interp(config.iac_cold_idle_temp,
		config.iac_warm_idle_temp,
		config.iac_cold_rpm,
		config.iac_warm_rpm,
		engine.coolant);
#ifdef A340E
	if(at_control.at_status&_BV(IDLE_UP))
		final_rpm+=at_config.idle_rpm_up;
#endif
	return final_rpm;
}

/** =================================
* lookup "iac-should-be-around-this-position-at-this-temperature" position
* Higher refpos as MAT goes from 0C to 70C (max +13%, that is +32 steps)
* 0C and 70C:
*/
#define IACREF_MATMIN 100
#define IACREF_MATMAX 170

uint8_t lookup_iac_clt_ref_pos( void )
{
	uint8_t t8, matslope;
	t8 = linear_interp(warmup_clt_range(clt_idx.lindex),
		warmup_clt_range(clt_idx.uindex),
		config.iac_ref_pos[clt_idx.lindex],
		config.iac_ref_pos[clt_idx.uindex],
		engine.coolant);

	matslope = config.iacref_matslope & 0xF;
	if( matslope ){
		uint16_t t16;
		uint8_t mat = engine.mat;
		/* mat -= IACREF_MATMIN but no overflow: */
		if ( mat <= IACREF_MATMIN ){
			mat = 0;
		} else if (mat < IACREF_MATMAX ) {
			mat -= IACREF_MATMIN;
		} else {
			mat = IACREF_MATMAX - IACREF_MATMIN;
		}
		t16 = mat * matslope;
		t16 >>= 5; // 70*15 /32 = max +32.8 (+32 steps = +12.8%)
		t16 += t8;
		if( t16 >> 8 ){
			t8 = 0xFF;
		} else {
			t8 = t16 & 0xFF;
		}
	}
#ifdef A340E
	if(at_control.at_status&_BV(IDLE_UP))
		t8+=at_config.idle_ref_up;
#endif
	return t8;
}

/**
* Freakin Fantastic Idle Control [FFIC (tm)]
* proportional change in ignition advance as function of deviance from target rpm.
* adv must be IGNADV_LOCALOFFSET based (see ignadv_table_lookup() )
* so underflow check not needed here.
*/
uint16_t idle_advance_adjust(uint16_t adv) {
	uint16_t new_adv;

	new_adv = adv;
	if ((config.iac_conf & _BV(advance_control)) && (engine.status & _BV(idle))) {
		uint16_t current_rpm, target_rpm;
		uint8_t limit;

		current_rpm = engine.rpm_vhr;
		
		// rpm is way to high, we prepare the advance idle mode but no adjustments yet
		// we dont want the pid (or advance controller for that matter) to start acting before we get in the 'target zone'
		if (engine.rpm > config.overrun_fuelresume) {	
			return adv;
		}
		
		target_rpm = iac.target_rpm * 10;
		
		if (current_rpm < target_rpm) {  // going too slow; more advance
			uint16_t adjust, t;
			t = target_rpm - current_rpm;

			adjust = mult16_8(t, config.iac_ign_slope) >> 8;
			limit = config.iac_ign_advance_limit;

			if (adjust > limit)
				adjust = limit;

			// overflow uint16_t is NOT possible,
			// but stupid configuration like ~50 deg ignadv
			// at RPM lower than idle and +14 or more degrees here,
			// result can be > 0x01FF which might act strange
			new_adv += adjust;
		} else {  // going too fast; less advance
			uint16_t adjust, t;
			t = current_rpm - target_rpm;

			adjust = mult16_8(t, config.iac_ign_slope) >> 8;
			limit = config.iac_ign_retard_limit;

			if (adjust > limit)
				adjust = limit;

			new_adv -= adjust;
		}
	}

	iac.adv_adjust = new_adv - adv;
	return new_adv;
}

/**
* Call this appr every 100 msec.
* Allows quick change up,
* but limits the change in the down direction (exponential decay)
* current_rpm is [1 RPM] unit
* final_target is [10 RPM] unit
* returns new iac.rpm_target (and saves it too)
*/
uint8_t graceful_decay(uint8_t final_target, uint16_t current_rpm)
{
	uint16_t t16;
	uint8_t old_target, new_target;
	// (22*final_target + 10*current) /32
	// == (11*final_target + (10*current)/2) /16	
	t16 = 11 * final_target;
	// current_rpm is already (10*current), because of high precision
	t16 += (current_rpm >> 1); // overflow not possible
	t16 >>= 4;
	new_target = (t16 & 0xFF00) ? 0xFF : (uint8_t)t16;

	if( new_target <= final_target){
		// as new_target is a weighting of final_target and current:
		// this can only happen if current_rpm <= final_target*10 
		// don't allow target going under final_target
		new_target = final_target;
	} else if( new_target < (old_target = iac.target_rpm) ){		
		// old_target must converge to final_target: new = (15*old + final)/16
		// (15/16) ** 10 = 0.524
		// so half-life time is appr 1 sec (10*100 msec)
		t16 = old_target * 15;
		t16 += final_target;
		t16 >>= 4;
		new_target = (t16 & 0xFF00) ? 0xFF : (uint8_t)t16;
	} // else we already have the new target
	iac.target_rpm = new_target;
	return new_target;
}

/**
 * Originally for some RPM kick at startup to kickstart the alternator.
 * More useful when reaching iac_integral_limit_inc (magic AC-ON detected)
 */
void afterstart_kick( void )
{
	iac.afterstart_duration = config.iac_afterstart_duration;
	iac.afterstart_rpm = config.iac_afterstart_rpm;
}

/**
 * magic-ac hack.
 * iac_afterstart_duration, iac_afterstart_rpm, iac_cold_rpm, iac_warm_rpm all odd (not even),
 * and iac_afterstart_rpm is 11 to 31 (added RPM is 110 to 310) 
 * TODO: find a separate config bit for it ?
 */
uint8_t magic_ac_configured( void )
{
	uint8_t odd = config.iac_afterstart_duration;
	odd &= config.iac_afterstart_rpm;
	odd &= config.iac_cold_rpm;
	odd &= config.iac_warm_rpm;

	if( odd & 1 ){ // wow, everything is configured odd
		odd = config.iac_afterstart_rpm;
		if( odd < 11 || odd > 31){
			return 0;	// no magic
		} else {
			return 1;	// yes, everything is found => magic_ac !
		}
	} else {	// some variable is even, no magic !
		return 0; 
	}
}

/** "Afterstart" Added Air
* decreases iac.afterstart_rpm (unless already 0)
*/
void afterstart_update( void )
{
	if( iac.afterstart_rpm == 0 ){
		return;	// nothing to do
	}

	if( magic_ac_configured() ){
	/* new triangle-behaviour:
	 * afterstart_rpm drops 10 RPM every iac_afterstart_duration [0.1sec]
	*/
		if (iac.afterstart_duration){
			iac.afterstart_duration--;
		} else {
			iac.afterstart_duration = config.iac_afterstart_duration;
			iac.afterstart_rpm--;
		}
	} else {	/* originally trapezoid behavior:
	 	* first afterstart_duration decays to 0, than afterstart_rpm quickly decays to 0:
		*/
		if (iac.afterstart_duration) iac.afterstart_duration--;
		else iac.afterstart_rpm--;
	}
}

/** =================================
 * (int16_t)iac.integral += iac_error
 * but respect integral limits. A bit of a bithacking as you see.
 * Internally we use 0x8000 based (uint16_t) numbers
 * This might be one of the deepest stack-users.
 * together with calc_iac_pos(), appr 16 bytes stack used (+add main())
 * Function could return if iac_integral_limit_inc was reached.
 */
void iac_integral( int16_t iac_error )
{
	uint32_t u32;
	uint16_t iac_integral, limit;
	uint8_t u8;
	// integral windup
	if ((iac.position == 0) && (iac_error < 0)) {
		// negative error, but iac can't move any lower
		return;
	}
	if ((iac.position >= config.iac_max_steps) && (iac_error > 0)) {
		// positive error, but iac can't move any higher
		return;
	}

	// integrate error term
	iac_integral = (uint16_t) iac.integral;
	iac_integral ^= 0x8000;	// -32768 .. 0 moved to 0 .. 0x8000
	u32 = iac_integral;
	iac_integral = (uint16_t) iac_error; // temporarily ruin iac_integral
	iac_integral ^= 0x8000;
	u32 += iac_integral;	// 0x10000 based
	u32 += 0x00008000;	// 0x18000 based
	iac_integral = u32;	// save before u32 ruined
	u32 >>= 16;
	u8 = u32;	// u8 == 0 if underflow, 2 if overflow, 1 normally
	if (u8 == 0)
		iac_integral = 0;
	else if (u8 >= 2)
		iac_integral = 0xFFFF;

	// limit the integrator
	limit = 0x8000;
	limit -= ((uint16_t)config.iac_integral_limit_dec << INTEGRAL_LIMIT_SCALE);
	if( iac_integral < limit ){	// lower limit
		iac_integral = limit;
	} else {
		limit = (uint16_t)config.iac_integral_limit_inc << INTEGRAL_LIMIT_SCALE;
		limit += 0x8000;
		if( iac_integral > limit ){	// upper limit
			// normally integral is around limit_dec; This is only reached
			// with air-conditioner and bigfan (maybe rear-window-heater?)

			if( magic_ac_configured() ){
				// Note that this will effect next calc_iac_pos() 0.1 sec later
				afterstart_kick();
				// We could return true to apply higher target early;
			}
			iac_integral = limit;
		}
	}

	iac.integral = iac_integral ^ 0x8000; // change back to int16_t
}

/**
 * returns error * scale/16
 * clamps result to max 0x7FFF
 * ONLY call for non-negative numbers !
 */
uint16_t scale_error( uint16_t error, uint8_t scale )
{
	uint32_t u32 = mult16_8( error, scale );
	u32 >>= 4;
	if( u32 & 0xFF8000){	// u32 >= 8000, so clamp
		return 0x7FFF;
	}
	return (uint16_t) u32;
}

/**
 * Caller passes: iac_clt = lookup_iac_clt_ref_pos();
 */
uint8_t tps_follow_position( uint8_t iac_clt )
{
	if(iac_tps_follow_configured()) { // throttle follow mode:
		// interpolate between iac_clt_reference position and max iac pos

		uint8_t iac_tps;
		iac_tps = actual_tps_thres;

/* Do NOT raise for iac_tps_thres + 7% (18/256), where the throttle is too sensitive anyway
 * actually, closing iac while TPS goes from 0..7% would make sense to make TPS less
 * sensitive in this region. Unfortunately Bosch iac requires 25% duty to close (not 0%)
 * so we leave this for now
 */
		if( iac_tps >= 255 - NOEXTRAAIR_TPSWINDOW){	//#!what is this for?
			return iac_clt;
		} 

//#1.1 interpolate between iac.last_pos and iac_pos(iac_clt), preventing transients when leaving idle mode

		if (engine.tps <= iac_tps + NOEXTRAAIR_TPSWINDOW) {
			iac_tps = linear_interp(iac_tps, (iac_tps + NOEXTRAAIR_TPSWINDOW), iac.last_pos, iac_clt, engine.tps);
		} else {
			iac_tps += NOEXTRAAIR_TPSWINDOW;
			iac_tps = linear_interp(iac_tps, 255, iac_clt, config.iac_max_steps, engine.tps);
			iac.last_pos = iac_tps;		//#1.1 preventing transient when tps entering idle again
		}

		return iac_tps;

	} else {
		return 0;
	}
}

/**
 * iac.integral is used, but NOT updated
 * caller must call iac_integral( iac_error ) beforehand if he likes
 * caller must fill iac.rpmdot = iac.last_rpm - current_rpm;
 */
void calc_iac_pid( int16_t iac_error )
{
	int16_t iac_pid;

	// asymmetric behavior. If rpm < target, increase actuation fast
	// bump up P and D-falling term, but not D-rising and I term.
	// I term is unsuitable for fast action anyway, would just wind up intgegal
	if (config_iac_pid_asymetric() && iac_error > 0) {
		uint8_t scale;
		if( iac_error < 48){
			scale = 16 + iac_error;	// scale *1.0 .. 3.9375
		} else { // iac_error is 48 RPM or higher,
			scale = 64;	// scale very strong (*4.0)
		}
		iac_error = scale_error( iac_error, scale );
		if( iac.rpmdot > 0 ){ // only scale up D for falling RPM
			iac.rpmdot = scale_error( iac.rpmdot, scale);
		}
	}

	iac.pid_p = mult16i_8_div256(iac_error, config.iac_kp);
	iac.pid_i = mult16i_8_div256(iac.integral, config.iac_ki);
	iac.pid_d = mult16i_8_div256(iac.rpmdot, config.iac_kd);

	// TODO: hope this doesn't overflow
	// with sufficiently small (iac_kp + iac_ki + iac_kd <= 256) values will never overflow
	// but in practice iac_error and iac.rpmdot is also small (compared to 32767 RPM)
	// and iac_integral_limit helps too => no worries
	iac_pid = iac.pid_p + iac.pid_i + iac.pid_d;

	// divide by 16 to allow larger constants in the above 
	// iac.pid = mult16i_8_div256(iac.pid, 16);
	iac.pid = div16i_exp2(iac_pid, 4);
}

/** =========================================
* NEW calc iac pos
* Why don't we use pid15() + iac_ref_pos(clt)
* after finding out iac target RPM ?
* The unit of internal rpm variables is [1 rpm] unless otherwise noted
* Caller should only call if(config.iac_conf & _BV(precise_idle))
* See /svn/firmware/firmware/trunk/doc/iac/iac_operation.* (PWM iac)
* and the stepper is simpler (no motivation to close it)
*/
void calc_iac_pos(void) {
	uint16_t current_rpm;	// 1 RPM
	int16_t iac_error;	// 1 RPM, positive below RPM target
	uint8_t final_rpm;	// final target RPM [10 RPM]
	uint8_t iac_clt;	// CLT reference position
	uint8_t pos;	// final iac.position
	uint8_t u8;// flags PID 100% authenticity (with integral, and no twilight zone gradient)

#ifdef MWFIRE_HIJACKED_IAC
	// MWfire: testing stepper motor to control LCS lpg system
	current_rpm = inj_port[0].pwcalc; // 4 usec
	current_rpm >>= 4;		// => 64 usec 
	pos = current_rpm >= 256 ? 0xFF : current_rpm & 0xFF;	// no overflow when pw > 16384 usec
	set_iac_pos(pos);	// 0 .. 100% according to injpw 0..16.4 msec
	return;
#endif

	if ((!(engine.status & _BV(running))) || (engine.status & _BV(crank))) { // engine ain't running or in cranking mode
		uint8_t cranking_pos;
		cli();
		engine.status &= ~_BV(idle);
		sei();

		// wind up integral: #1.1 no!!, deleting..
		iac.integral = 0; 
		iac.rpmdot = 0;
		iac.pid = 0;
		iac.pid_p = 0;
		iac.pid_i = 0;
		iac.pid_d = 0;
//		iac_resync_reset();
		iac.closing_delay = config.iac_close_delay * 10; //#1.1
		iac.adv_adjust = 0;

		// park the iac at the starting position
		afterstart_kick();
		iac.target_rpm = add8_8( config.iac_cold_rpm, iac.afterstart_rpm ); 

		if( iac_stepper_configured()	// no 0 command to stepper iac !
			|| (engine.status & _BV(crank))
			|| (etime.now_upper < config.pump_on_mintime) ) {
			// cranking position after bootup for a bit shorter than fuelpump
			// ( using "< pump_on_mintime" instead of "<= pump_on_mintime" )
			// and during cranking of course
			cranking_pos = linear_interp( COLD_TEMP, WARM_TEMP, config.iac_cold_start_pos, config.iac_warm_start_pos, engine.coolant );
		} else { // solenoid duty will be 0 (even after VBATT correction)
			cranking_pos = 0;
		}
		set_iac_pos(cranking_pos);
		return;
	}

//engine is running

	current_rpm = engine.rpm_vhr;  		// only written from userspace, no need for cli/sei
								// rising RPM => negative rpmdot ('error' defined this way)
	iac.rpmdot = iac.last_rpm - current_rpm;	// before ruining iac.last_rpm
	iac.last_rpm = current_rpm;			// rpmdot calculation will need it
	iac_clt = lookup_iac_clt_ref_pos();

	// ---------- target RPM calculation: ------------
	// engine.rpm==9, REAL_RPM == rpm_vhr() == 900 
	// for target config.idle_rpm == 90 == 0x5a
	final_rpm = lookup_clt_rpmtarget();
	afterstart_update();
	final_rpm = add8_8( final_rpm, iac.afterstart_rpm );
	// final_rpm => current target (the name stays)
	final_rpm = graceful_decay(final_rpm, current_rpm);

	// ------------ PID or PD calculation: -----------
	// error term: positive error means RPM lower than supposed
	iac_error = (final_rpm * 10) - current_rpm;

// tps~0 or RPM below target
	if( engine.tps <= actual_tps_thres || iac_error > 0 ){ 
		u8 = 1;
		// within integral deadband, prevent stepper adjusting all the time
		uint16_t abs_error = abs(iac_error);


		uint8_t integral_kpa = engine_kpa8();

//#1.1 low-map integral shut-down. rpm over overrun_fuelcut, or under user-defined kPa (engine braking)
		if ((integral_kpa < config.iac_integral_lowmap) || (engine.rpm > config.overrun_fuelresume)) {	
			iac.integral = 0;		// pos = iac_clt + PD
			iac.status |= _BV(no_integral); 	// no integral control!

// map-reading over user-defined kPa + offset, and rpm below overrun_fuelcut
		} else if ((iac.status & _BV(no_integral))&&
			((config.iac_integral_lowmap + engine_kpa8_of(16)) < integral_kpa)) {	// offset 4kpa
			iac.status &= ~_BV(no_integral);	// use integral control!
		}

// don't update integral if bit is set, or when error is too small
		if ((abs_error >= config.iac_integral_deadband) && !(iac.status & _BV(no_integral))) {
			iac_integral( iac_error );
		}

//not in idle
	} else { // do NOT update integral to avoid wind-to-negative.
		u8 = 0;
		iac.status &= ~_BV(no_integral);	// unset 'no_integral', just for safety...
	}

//---#

	calc_iac_pid( iac_error );
	pos = add8_16i(iac_clt, iac.pid);

	if (u8) { 				// PID-controlled idle-mode
		in_idle();
		iac.last_pos = pos;	//#1.1
		set_iac_pos(pos);
		return;
	} else {
		// not in idle-mode...
		// TWILIGHT zone. solenoid somewhere between last PID-calculated position, and 0 duty
		// or between last PID-calculated position, and tps_fol (either stepper and solenoid)

		pos = tps_follow_position( iac_clt );
		not_in_idle();	//close solenoid, and zero integral after a while if NOT "tps_follow" configured.
					//preventing rpm drops(because integral=0) if just pressed throttle for a short time..
		if (pos) {	//zero if tps_follow_mode not configured
			set_iac_pos(pos);
		}
	}
}

/**
 * Set iac.position to pos (but max config.iac_max_steps)
 * TODO: do we need cli()-sei() protection for iac.status ?
 * TODO: Test: now we return immediately if pos == iac.position ? (for efficiency)
 * TODO: uint16_t pos ??
 * Applies VBATT correction for PWM iac
 * iac.position can be used as a base calculation for
 * next pos. Not needed normally, as it's calculated from scratch using PID integral,
 * but can be useful, as in slowly_close_solenoid() or pos +- 1 inside deadband
 */
void set_iac_pos(uint8_t pos)
{
	uint8_t oldposition;

	// rail to config.iac_max_steps:
	if( pos > config.iac_max_steps ) pos = config.iac_max_steps;

	oldposition = iac.position;

	if(iac_stepper_configured()) {
/*		if( pos == 0 && stepper_time_to_overclose() ) {
			// time for overclose: cheat the iac, it will think oldposition was 1
			oldposition = 1;
		}*/
		if (pos == oldposition || iac.status & _BV(sync)) {   // don't mess with the sync sequence!
			return;
		}
		uint8_t count;
		uint8_t status = iac.status;
		if (pos < oldposition) {   // close iac
			count = oldposition - pos;
			status &= ~_BV(direction);
			status |= _BV(busy);
		} else {   // open iac
		// when pos == oldposition, status became busy with count=0
		// Now changed to do nothing, hopefully not broke anything. 
			count = pos - oldposition;
			status |= _BV(direction) | _BV(busy);
		}
		iac.status = status;
		iac.shared.stepperonly.count = count;
	} else {	// PWM iac
		iac.position = pos;	// save new position (not duty for solenoid)
		uint8_t duty = calc_pwm_dc( pos ); // apply VBATT scaling
		if( duty > config.iac_max_steps ) duty = config.iac_max_steps;
		// duty might be the same as earlier. Who cares, this is very cheap:
		set_iac_solenoid_duty( duty );
	}
}



/***************************************************************************/
/*** init iac                                                            ***/
/***************************************************************************/
void init_iac(void) {
	if(iac_solenoid_configured()) { // solenoid iac: PWM or ON/OFF
#ifndef TRUE_PWM_IAC // new experimental
		softpwm_init(&iac.shared.idle_pwm_soft, 0);
#endif
		iac.status &= ~(_BV(busy) | _BV(sync));
		iac.position = 0;
	} else { // stepper IAC
		iac.status = _BV(busy) | _BV(sync);
		iac.status &= ~_BV(direction);
		iac.position = config.iac_max_steps + IAC_SYNC;
		iac.shared.stepperonly.count = config.iac_max_steps + IAC_SYNC;
	}
	// we always enable the chip, NOT just when stepper enabled
	// So ignout or other outputs can use it.
	cli();
	S259(DHIGH, S259_STEP_EN); 
	sei();
}

/***************************************************************************/
/*** move iac                                                            ***/
/*** caller MUST check if stepper_iac is enabled						 ***/
/***************************************************************************/

void move_iac(void) {
	if (iac.shared.stepperonly.count) {

		uint8_t next_phase;

		iac.shared.stepperonly.count--;

		if (iac.status & _BV(direction)) {	// open iac
			next_phase = iac.shared.stepperonly.phase + 1; 
			iac.position++; // phase and position does not represent the same number!
		} else {	// close iac
			next_phase = iac.shared.stepperonly.phase - 1; 
			iac.position--;
		}

		if (config_ext.iac_conf2&_BV(toyota_stepper))//toyota stepper at p259 4-7
		{
			cli();
			P259(DHIGH, next_step(next_phase & 0x03)+1);  // activate next wire in sequence
			sei_nop_cli();
			P259(DLOW, next_step(iac.shared.stepperonly.phase & 0x03)+1);  // deactivate previous wire in sequence
		}
		else
		{
			if (config_ext.iac_conf2&_BV(stepper_negative))//negative
			{
				cli();
			//	S259(DLOW, next_step(next_phase & 0x03));  // activate next wire in sequence
				S259(DHIGH, next_step(iac.shared.stepperonly.phase & 0x03));  // deactivate previous wire in sequence
				sei_nop_cli();
			//	S259(DHIGH, next_step(iac.shared.stepperonly.phase & 0x03));  // deactivate previous wire in sequence
				S259(DLOW, next_step(next_phase & 0x03));  // activate next wire in sequence
			}
			else //old positive
			{
				cli();
		//		S259(DHIGH, next_step(next_phase & 0x03));  // activate next wire in sequence
				S259(DLOW, next_step(iac.shared.stepperonly.phase & 0x03));  // deactivate previous wire in sequence
				sei_nop_cli();
		//		S259(DLOW, next_step(iac.shared.stepperonly.phase & 0x03));  // deactivate previous wire in sequence
				S259(DHIGH, next_step(next_phase & 0x03));  // activate next wire in sequence
			}
		}

	//	sei_nop_cli();

	//	S259(DHIGH, S259_STEP_EN); // this is a bit waste, TODO: find better place to enable the driver
	//	sei();

		iac.shared.stepperonly.phase = next_phase;

	//	if (iac.shared.stepperonly.count == 0) {
	//		iac.status &= ~(_BV(busy) | _BV(sync));  // when done sync'ing, the position is know
	} else {  // no step requested, disable driver
		iac.status &= ~(_BV(busy) | _BV(sync));
		if (config.iac_conf & _BV(power_off_iac)) {
			cli();
			S259(DLOW, S259_STEP_EN);
			sei_nop_cli();
			S259(DLOW, S259_STEP_A);
			sei_nop_cli();
			S259(DLOW, S259_STEP_B);
			sei_nop_cli();
			S259(DLOW, S259_STEP_C);
			sei_nop_cli();
			S259(DLOW, S259_STEP_D);
			sei();
		}
	}
}


uint8_t PROGMEM STEP_PINS[] = { S259_STEP_A, S259_STEP_B, S259_STEP_C, S259_STEP_D };


/***************************************************************************/
/*** next_step                                                           ***/
/***************************************************************************/
// Allow allmost total reconfiguration of the steppers windings connections 
// through PC configuration program. Nice for people too lazy for soldering the
// stepper wires correctly!
uint8_t next_step(uint8_t phase) {
  uint8_t val, idx;

  switch (phase) {
  case 0: idx = config.iac_step_seq & 0x03; break;
  case 1: idx = (config.iac_step_seq >> 2) & 0x03; break;
  case 2: idx = (config.iac_step_seq >> 4) & 0x03; break;
  case 3: idx = (config.iac_step_seq >> 6); break;
  default: idx = 0; break;
  }
  
  val = PRG_RDB(&STEP_PINS[idx]);
  return val;
}


/** ONLY call from userspace!
 * If iac_pwm_configured() we use PWM, otherwise ON/OFF
 * Check digitalout() for how to configure config.iac_sol_channel
 * so DISABLED, p259, s259, i259, stepper or FET can be chosen
 */
void idle_solenoid(void)
{
	uint8_t output;
	uint8_t sol_channel = config.iac_sol_channel;
	if( digitalout_disabled( sol_channel ) ){
		return;
	}

	if( iac_pwm_configured() ) {
#ifdef TRUE_PWM_IAC // new experimental
		if (mainlppwm.pwm_states&_BV(ML_PWM_IAC)) //rising edge
		{
			output=DHIGH;
			mainlppwm.pwm_states&=~_BV(ML_PWM_IAC); //next is falling
		}
		else // falling edge
		{
			output=DLOW;
			mainlppwm.pwm_states|=_BV(ML_PWM_IAC); //next is rising
		}
#else
		output = softpwm_act(&iac.shared.idle_pwm_soft);
#endif
	} else {
/** fast idle == on/off solenoid to raise idleRPM when cold.
 * FAST_IDLE_THRES avoids erratic idle when coolant == fastidle
 * TODO: drop fastidle variable: reuse iac_cold_rpm and iac_warm_rpm (for hysteresis thresholds)
 */
		if (engine.coolant >= config.iac_warm_idle_temp) {
			output = 0;
		} else if (engine.coolant < config.iac_cold_idle_temp){
			output = DHIGH;		
		} else { // hysteresis window, don't touch the output
			// if rebooted in the hysteresis window, output can be ON/OFF - uncertain
			return;
		}
	} // if not precise_idle

	// turns on idle_output if state==DHIGH.
	// turns off idle_output if state==0
	cli();
	digitalout( sol_channel, output);
	if( iac_dualpwm_configured() ) {
		sei_nop_cli();
        	digitalout( sol_channel + 0x10, output ^ 0x80 );
	}
	sei();
} // idle_solenoid()

/*@}*/
